#include "io/model/Mesh.h"

namespace ehs
{
	Mesh::Mesh()
    {
		AddType("Mesh");
    }

	Mesh::Mesh(Str_8 id, Array<Vertex_f> vertices, Array<UInt_32> indices)
        : Resource(std::move(id)), vertices(std::move(vertices)), indices(std::move(indices))
    {
		AddType("Mesh");
    }

	Mesh::Mesh(Str_8 id, Array<Vertex_f> vertices)
        : Resource(std::move(id)), vertices(std::move(vertices))
    {
		AddType("Mesh");
    }

	Mesh::Mesh(Str_8 id)
        : Resource(std::move(id))
    {
		AddType("Mesh");
    }

	Mesh::Mesh(Mesh&& mesh) noexcept
		: Resource(std::move(mesh)), vertices(std::move(mesh.vertices)),
		indices(std::move(mesh.indices))
	{
	}

	Mesh::Mesh(const Mesh& mesh)
		: Resource(mesh), vertices(mesh.vertices), indices(mesh.indices), srcVertBuffer(mesh.srcVertBuffer),
		  dstVertBuffer(mesh.dstVertBuffer), srcIndBuffer(mesh.srcIndBuffer), dstIndBuffer(mesh.dstIndBuffer)
	{
	}

	Mesh& Mesh::operator=(Mesh&& mesh) noexcept
	{
		if (this == &mesh)
			return *this;

		vertices = std::move(mesh.vertices);
		indices = std::move(mesh.indices);
		srcVertBuffer = std::move(mesh.srcVertBuffer);
		dstVertBuffer = std::move(mesh.dstVertBuffer);
		srcIndBuffer = std::move(mesh.srcIndBuffer);
		dstIndBuffer = std::move(mesh.dstIndBuffer);

		Resource::operator=(std::move(mesh));

		return *this;
	}

	Mesh& Mesh::operator=(const Mesh& mesh)
	{
		if (this == &mesh)
			return *this;

		vertices = mesh.vertices;
		indices = mesh.indices;
		srcVertBuffer = mesh.srcVertBuffer;
		dstVertBuffer = mesh.dstVertBuffer;
		srcIndBuffer = mesh.srcIndBuffer;
		dstIndBuffer = mesh.dstIndBuffer;

		Resource::operator=(mesh);

		return *this;
	}

	bool Mesh::UploadToGpu(GpuCmdBuffer* cmdBuffer)
	{
		GameLoop* gl = (GameLoop*)GetParent("GameLoop");
		if (!gl)
			return false;

		RenderWindow* win = gl->GetWindow();
		if (!win)
			return false;

		GpuInterface* inf = win->GetInterface();

		if (readOnly)
		{
			srcVertBuffer = {
				inf,
				VK_BUFFER_USAGE_TRANSFER_SRC_BIT,
				sizeof(Vertex_f) * vertices.Size()
			};

			Byte* data = (Byte*)srcVertBuffer.MapMemory();

			for (UInt_64 i = 0; i < vertices.Size(); ++i)
				for (UInt_64 v = 0; v < sizeof(Vertex_f); ++v)
					data[sizeof(Vertex_f) * i + v] = ((Byte*) &vertices[i])[v];

			srcVertBuffer.UnMapMemory();

			dstVertBuffer = {
				inf,
				VK_BUFFER_USAGE_TRANSFER_DST_BIT | VK_BUFFER_USAGE_VERTEX_BUFFER_BIT,
				srcVertBuffer.Size(),
				VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT
			};

			VK_MEMORY_PROPERTY_HOST

			VkBufferMemoryBarrier barrier = {};
			barrier.sType = VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER;
			barrier.pNext = nullptr;
			barrier.srcAccessMask = 0;
			barrier.dstAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
			barrier.srcQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
			barrier.dstQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
			barrier.buffer = dstVertBuffer;
			barrier.offset = 0;
			barrier.size = VK_WHOLE_SIZE;

			cmdBuffer->BufferBarrier(
				VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT,
				VK_PIPELINE_STAGE_TRANSFER_BIT,
				0,
				1,
				&barrier
			);

			GpuBuffer::Copy(cmdBuffer, &srcVertBuffer, &dstVertBuffer);

			barrier.sType = VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER;
			barrier.pNext = nullptr;
			barrier.srcAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
			barrier.dstAccessMask = VK_ACCESS_SHADER_READ_BIT;
			barrier.srcQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
			barrier.dstQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
			barrier.buffer = dstVertBuffer;
			barrier.offset = 0;
			barrier.size = VK_WHOLE_SIZE;

			cmdBuffer->BufferBarrier(
				VK_PIPELINE_STAGE_TRANSFER_BIT,
				VK_PIPELINE_STAGE_VERTEX_SHADER_BIT,
				0,
				1,
				&barrier
			);

			if (indices.Size())
			{
				srcIndBuffer = {
					inf,
					VK_BUFFER_USAGE_TRANSFER_SRC_BIT,
					sizeof(UInt_32) * indices.Size()
				};

				UInt_32* indMem = (UInt_32*)srcIndBuffer.MapMemory();

				for (UInt_64 i = 0; i < indices.Size(); ++i)
					indMem[i] = indices[i];

				srcIndBuffer.UnMapMemory();

				dstIndBuffer = {
					inf,
					VK_BUFFER_USAGE_TRANSFER_DST_BIT | VK_BUFFER_USAGE_INDEX_BUFFER_BIT,
					srcIndBuffer.Size(),
					VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT
				};

				barrier.sType = VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER;
				barrier.pNext = nullptr;
				barrier.srcAccessMask = 0;
				barrier.dstAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
				barrier.srcQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
				barrier.dstQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
				barrier.buffer = dstIndBuffer;
				barrier.offset = 0;
				barrier.size = VK_WHOLE_SIZE;

				cmdBuffer->BufferBarrier(
					VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT,
					VK_PIPELINE_STAGE_TRANSFER_BIT,
					0,
					1,
					&barrier
				);

				GpuBuffer::Copy(cmdBuffer, &srcIndBuffer, &dstIndBuffer);

				barrier.sType = VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER;
				barrier.pNext = nullptr;
				barrier.srcAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
				barrier.dstAccessMask = VK_ACCESS_SHADER_READ_BIT;
				barrier.srcQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
				barrier.dstQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
				barrier.buffer = dstIndBuffer;
				barrier.offset = 0;
				barrier.size = VK_WHOLE_SIZE;

				cmdBuffer->BufferBarrier(
					VK_PIPELINE_STAGE_TRANSFER_BIT,
					VK_PIPELINE_STAGE_VERTEX_SHADER_BIT,
					0,
					1,
					&barrier
				);
			}

			cmdBuffer->AddDependency(this);
		}
		else
		{
			dstVertBuffer = {
				inf,
				VK_BUFFER_USAGE_VERTEX_BUFFER_BIT,
				sizeof(Vertex_f) * vertices.Size()
			};

			void* data = dstVertBuffer.MapMemory();
			Util::Copy(data, &vertices[0], dstVertBuffer.Size());
			dstVertBuffer.UnMapMemory();

			if (indices.Size())
			{
				dstIndBuffer = {
					inf,
					VK_BUFFER_USAGE_INDEX_BUFFER_BIT,
					sizeof(UInt_32) * indices.Size()
				};

				data = dstIndBuffer.MapMemory();
				Util::Copy(data, &indices[0], dstIndBuffer.Size());
				dstIndBuffer.UnMapMemory();
			}
		}

		return true;
	}

	bool Mesh::PostGpuUpload()
	{
		srcVertBuffer.Release();
		srcIndBuffer.Release();

		return true;
	}

	bool Mesh::HasPostGpuUploaded() const
	{
		return !srcVertBuffer.IsValid() && !srcIndBuffer.IsValid();
	}

	bool Mesh::ReleaseFromGpu()
	{
		srcVertBuffer.Release();
		srcIndBuffer.Release();
		dstVertBuffer.Release();
		dstIndBuffer.Release();

		return true;
	}

	bool Mesh::IsUploaded() const
	{
		return dstVertBuffer.IsValid();
	}

	void Mesh::Bind(GpuCmdBuffer* cmdBuffer)
	{
		cmdBuffer->AddDependency(this);

		if (dstIndBuffer.IsValid())
		{
			cmdBuffer->AddDependency(&dstIndBuffer);
			vkCmdBindIndexBuffer(*cmdBuffer, dstIndBuffer.GetBuffer(), 0, VK_INDEX_TYPE_UINT32);
		}

		cmdBuffer->AddDependency(&dstVertBuffer);

		VkDeviceSize offset = 0;
		VkBuffer vertB = dstVertBuffer.GetBuffer();
		vkCmdBindVertexBuffers(*cmdBuffer, 0, 1, &vertB, &offset);
	}

	void Mesh::Draw(GpuCmdBuffer* cmdBuffer, const UInt_32 instances)
	{
		if (dstIndBuffer.IsValid())
			vkCmdDrawIndexed(*cmdBuffer, indices.Size(), instances, 0, 0, 0);
		else
			vkCmdDraw(*cmdBuffer, vertices.Size(), instances, 0, 0);
	}

    void Mesh::SetVertices(const Array<Vertex_f>& newVertices)
    {
        vertices = newVertices;
    }

    Array<Vertex_f> Mesh::GetVertices() const
    {
        return vertices;
    }

    Array<Vertex_f>& Mesh::GetVertices()
    {
        return vertices;
    }

    void Mesh::SetIndices(const Array<UInt_32>& newIndices)
    {
        indices = newIndices;
    }

    bool Mesh::HasIndices() const
    {
        return indices.Size();
    }

    Array<UInt_32> Mesh::GetIndices() const
    {
        return indices;
    }

    Array<UInt_32>& Mesh::GetIndices()
    {
        return indices;
    }

	void Mesh::Calculate()
	{
		if (indices.Size())
			for (UInt_64 i = 0; i < indices.Size(); i += 3)
				Calculate(vertices[indices[i]], vertices[indices[i + 1]], vertices[indices[i + 2]]);
		else
			for (UInt_64 i = 0; i < vertices.Size(); i += 3)
				Calculate(vertices[i], vertices[i + 1], vertices[i + 2]);
	}

	void Mesh::Calculate(Vertex_f& vert1, Vertex_f& vert2, Vertex_f& vert3)
	{
		Vec3_f edge1 = vert2.pos - vert1.pos;
		Vec3_f edge2 = vert3.pos - vert1.pos;
		Vec2_f deltaUV1 = vert2.uv - vert1.uv;
		Vec2_f deltaUV2 = vert3.uv - vert1.uv;

		float f = 1.0f / (deltaUV1.x * deltaUV2.y - deltaUV2.x * deltaUV1.y);

		Vec3_f tan(
			f * (deltaUV2.y * edge1.x - deltaUV1.y * edge2.x),
			f * (deltaUV2.y * edge1.y - deltaUV1.y * edge2.y),
			f * (deltaUV2.y * edge1.z - deltaUV1.y * edge2.z)
		);

		vert1.tan = tan;
		vert2.tan = tan;
		vert3.tan = tan;

		Vec3_f bTan(
			f * (-deltaUV2.x * edge1.x + deltaUV1.x * edge2.x),
			f * (-deltaUV2.x * edge1.y + deltaUV1.x * edge2.y),
			f * (-deltaUV2.x * edge1.z + deltaUV1.x * edge2.z)
		);

		vert1.bTan = bTan;
		vert2.bTan = bTan;
		vert3.bTan = bTan;
	}
}